import os
import shutil
import sys

root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root_folder)
from ml4co_kit import TSPDataGenerator, MISDataGenerator, KaMISSolver


##############################################
#             Test Func For TSP              #
##############################################


def _test_tsp_lkh_generator(
    num_threads: int, nodes_num: int, data_type: str, 
    regret: bool, re_download: bool=False
):
    """
    Test TSPDataGenerator using LKH Solver
    """
    # save path
    save_path = f"tmp/tsp{nodes_num}_lkh"
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    # create TSPDataGenerator using lkh solver
    tsp_data_lkh = TSPDataGenerator(
        num_threads=num_threads,
        nodes_num=nodes_num,
        data_type=data_type,
        solver="lkh",
        train_samples_num=4,
        val_samples_num=4,
        test_samples_num=4,
        save_path=save_path,
        regret=regret,
    )
    if re_download:
        tsp_data_lkh.download_lkh()
    # generate data
    tsp_data_lkh.generate()
    # remove the save path
    shutil.rmtree(save_path)


def _test_tsp_concorde_generator(
    num_threads: int, nodes_num: int, data_type: str,
    recompile_concorde: bool = False
):
    """
    Test TSPDataGenerator using Concorde Solver
    """
    # save path
    save_path = f"tmp/tsp{nodes_num}_concorde"
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    # create TSPDataGenerator using lkh solver
    tsp_data_concorde = TSPDataGenerator(
        num_threads=num_threads,
        nodes_num=nodes_num,
        data_type=data_type,
        solver="concorde",
        train_samples_num=4,
        val_samples_num=4,
        test_samples_num=4,
        save_path=save_path,
    )
    if recompile_concorde:
        tsp_data_concorde.recompile_concorde()
        
    # generate data
    tsp_data_concorde.generate()
    # remove the save path
    shutil.rmtree(save_path)


def test_tsp():
    """
    Test TSPDataGenerator
    """
    _test_tsp_lkh_generator(
        num_threads=4, nodes_num=50, data_type="uniform", regret=False, re_download=True
    )
    _test_tsp_lkh_generator(
        num_threads=1, nodes_num=50, data_type="uniform", regret=True
    )
    _test_tsp_lkh_generator(
        num_threads=4, nodes_num=50, data_type="uniform", regret=True
    )
    _test_tsp_lkh_generator(
        num_threads=4, nodes_num=50, data_type="gaussian", regret=False
    )
    _test_tsp_concorde_generator(
        num_threads=4, nodes_num=50, data_type="uniform", recompile_concorde=True
    )
    _test_tsp_concorde_generator(
        num_threads=4, nodes_num=50, data_type="gaussian"
    )
    _test_tsp_concorde_generator(
        num_threads=4, nodes_num=50, data_type="cluster"
    )


##############################################
#             Test Func For MIS              #
##############################################


def _test_mis_kamis(
    nodes_num_min: int, nodes_num_max: int, data_type: str,
    recompile_kamis: bool = False
):
    """
    Test MISDataGenerator using KaMIS
    """
    # save path
    save_path = f"tmp/mis_{data_type}_kamis"
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    # create TSPDataGenerator using lkh solver
    solver = KaMISSolver(time_limit=10)
    if recompile_kamis:
        solver.recompile_kamis()
    mis_data_kamis = MISDataGenerator(
        nodes_num_min=nodes_num_min,
        nodes_num_max=nodes_num_max,
        data_type=data_type,
        solver=solver,
        train_samples_num=2,
        val_samples_num=2,
        test_samples_num=2,
        save_path=save_path,
    )
    # generate and solve data
    mis_data_kamis.generate()
    mis_data_kamis.solve()
    # remove the save path
    shutil.rmtree(save_path)


def _test_mis_gurobi(
    nodes_num_min: int, nodes_num_max: int, data_type: str
):
    """
    Test MISDataGenerator using MISGurobi
    """
    # save path
    save_path = f"tmp/mis_{data_type}_gurobi"
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    # create TSPDataGenerator using lkh solver
    mis_data_gurobi = MISDataGenerator(
        nodes_num_min=nodes_num_min,
        nodes_num_max=nodes_num_max,
        data_type=data_type,
        solver="gurobi",
        train_samples_num=2,
        val_samples_num=2,
        test_samples_num=2,
        save_path=save_path,
        solve_limit_time=10.0,
    )
    # generate and solve data
    mis_data_gurobi.generate()
    mis_data_gurobi.solve()
    # remove the save path
    shutil.rmtree(save_path)


def test_mis():
    """
    Test MISDataGenerator
    """
    _test_mis_kamis(
        nodes_num_min=600, nodes_num_max=700, data_type="er", recompile_kamis=True
    )
    _test_mis_kamis(nodes_num_min=600, nodes_num_max=700, data_type="ba")
    _test_mis_kamis(nodes_num_min=600, nodes_num_max=700, data_type="hk")
    _test_mis_kamis(nodes_num_min=600, nodes_num_max=700, data_type="ws")
    # gurobi need license
    # gurobipy.GurobiError: Model too large for size-limited license;
    # visit https://www.gurobi.com/free-trial for a full license
    # _test_mis_gurobi(nodes_num_min=600, nodes_num_max=700, data_type="er")
    # _test_mis_gurobi(nodes_num_min=600, nodes_num_max=700, data_type="ba")
    # _test_mis_gurobi(nodes_num_min=600, nodes_num_max=700, data_type="hk")
    # _test_mis_gurobi(nodes_num_min=600, nodes_num_max=700, data_type="ws")


##############################################
#                    MAIN                    #
##############################################

if __name__ == "__main__":
    test_tsp()
    test_mis()
    shutil.rmtree("tmp")



import os
import shutil
import pathlib


try:
    from .concorde.solve import TSPSolver as TSPConSolver
except:
    concorde_path = pathlib.Path(__file__).parent.parent / "pyconcorde"
    ori_dir = os.getcwd()
    os.chdir(concorde_path)
    os.system("python ./setup.py build_ext --inplace")
    os.chdir(ori_dir)
    if os.path.exists(f"{concorde_path}/build"):
        shutil.rmtree(f"{concorde_path}/build")
    from .concorde.solve import TSPSolver as TSPConSolver



import os
from ml4co_kit.utils import download, extract_archive


class SATLIBOriDataset:
    def __init__(self) -> None:
        self.url = "https://huggingface.co/datasets/ML4CO/SATLIBOriDataset/resolve/main/satlib_original.tar.gz?download=true"
        self.md5 = "0da8a73e2b79a6b5e6156005959ce509"
        self.dir = "dataset/satlib_original/"
        if not os.path.exists("dataset"):
            os.mkdir("dataset")
        if not os.path.exists(self.dir):
            download(filename="dataset/satlib_original.tar.gz", url=self.url, md5=self.md5)
            extract_archive(archive_path="dataset/satlib_original.tar.gz", extract_path=self.dir)



import os
import numpy as np
import pandas as pd
from ml4co_kit.data.tsp.tsplib_original import TSPLIBOriDataset
from ml4co_kit.solver.tsp.base import TSPSolver


class TSPLIBOriEvaluator:
    def __init__(self) -> None:
        self.dataset = TSPLIBOriDataset()
        self.support = self.dataset.support["resolved"]

    def evaluate(
        self,
        solver: TSPSolver,
        norm: str = "EUC_2D",
        normalize: bool = False,
        **solver_args
    ):
        # record
        solved_costs = dict()
        ref_costs = dict()
        gaps = dict()

        # get the evaluate files' dir and the problem name list
        evaluate_dir = self.support[norm]["path"]
        solution_dir = self.support[norm]["solution"]
        problem_list = self.support[norm]["problem"]

        # solve
        for problem in problem_list:
            # read
            file_path = os.path.join(evaluate_dir, problem + ".tsp")
            ref_tour_path = os.path.join(solution_dir, problem + ".opt.tour")
            solver.from_tsp(file_path, norm, normalize)
            solver.read_ref_tours_from_opt_tour(ref_tour_path)
            solver.solve(norm=norm, normalize=normalize, **solver_args)
            solved_cost, ref_cost, gap, _ = solver.evaluate(calculate_gap=True)
            # record
            solved_costs[problem] = solved_cost
            ref_costs[problem] = ref_cost
            gaps[problem] = gap

        # average
        np_solved_costs = np.array(list(solved_costs.values()))
        np_ref_costs = np.array(list(ref_costs.values()))
        np_gaps = np.array(list(gaps.values()))
        avg_solved_cost = np.average(np_solved_costs)
        avg_ref_cost = np.average(np_ref_costs)
        avg_gap = np.average(np_gaps)
        solved_costs["AVG"] = avg_solved_cost
        ref_costs["AVG"] = avg_ref_cost
        gaps["AVG"] = avg_gap

        # output
        return_dict = {
            "solved_costs": solved_costs,
            "ref_costs": ref_costs,
            "gaps": gaps,
        }
        df = pd.DataFrame(return_dict)
        return df



""" setup command to build Concorde Cython wrappers.

By default, the setup script will download the QSOpt linear solver, and
download and compile Concorde. If you have either one of these packages
already installed, you can use the installed version by setting the
following environment variables:

QSOPT_DIR: should point to a folder containing qsopt.a and qsopt.h
CONCORDE_DIR: contains concorde.a and concorde.h

Note that for the build process to work correctly, you should either
not set these variables (and rely on the downloaded Concorde) or set
both of them. Setting only one will not work as intended.

"""

from __future__ import print_function

import os
from os.path import exists, join as pjoin
import platform
import shutil
import subprocess

try:
    import urllib.request

    urlretrieve = urllib.request.urlretrieve
except ImportError:  # python 2
    from urllib import urlretrieve

from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

from Cython.Build import cythonize

import numpy as np

QSOPT_LOCATION = {
    "Darwin": {
        "arm64": (
            "https://www.math.uwaterloo.ca/~bico/qsopt/downloads/codes/m1/qsopt.a",
            "https://www.math.uwaterloo.ca/~bico/qsopt/downloads/codes/m1/qsopt.h",
        ),
        "x86_64": (
            "https://www.math.uwaterloo.ca/~bico/qsopt/downloads/codes/mac64/qsopt.a",
            "https://www.math.uwaterloo.ca/~bico/qsopt/downloads/codes/mac64/qsopt.h",
        ),
    },
    "Linux": {
        "x86_64": (
            "https://www.math.uwaterloo.ca/~bico/qsopt/beta/codes/PIC/qsopt.PIC.a",
            "https://www.math.uwaterloo.ca/~bico/qsopt/beta/codes/PIC/qsopt.h",
        ),
    },
}

CONCORDE_SRC = "https://www.math.uwaterloo.ca/tsp/concorde/downloads/codes/src/co031219.tgz"  # noqa


def _safe_makedirs(*paths):
    for path in paths:
        try:
            os.makedirs(path)
        except os.error:
            pass


def download_concorde_qsopt():
    _safe_makedirs("data")
    _safe_makedirs("build")
    qsopt_a_path = pjoin("data", "qsopt.a")
    qsopt_h_path = pjoin("data", "qsopt.h")
    if not exists(qsopt_a_path) or not exists(qsopt_h_path):
        print("qsopt is missing, downloading")
        machine = platform.machine()
        qsopt_a_url, qsopt_h_url = QSOPT_LOCATION[platform.system()][machine]
        urlretrieve(qsopt_a_url, qsopt_a_path)
        urlretrieve(qsopt_h_url, qsopt_h_path)
    concorde_src_path = pjoin("build", "concorde.tgz")
    if not exists(concorde_src_path):
        print("concorde is missing, downloading")
        urlretrieve(CONCORDE_SRC, concorde_src_path)


def _run(cmd, cwd):
    subprocess.check_call(cmd, shell=True, cwd=cwd)


def build_concorde():
    if not exists("data/concorde.h") or not exists("data/concorde.a"):
        print("building concorde")
        _run("tar xzvf concorde.tgz", "build")

        cflags = "-fPIC -O2 -g"

        if platform.system().startswith("Darwin"):
            flags = "--host=darwin"
        else:
            flags = ""

        datadir = os.path.abspath("data")
        cwd = (
            'CFLAGS="{cflags}" ./configure --prefix {data} '
            "--with-qsopt={data} {flags}"
        ).format(cflags=cflags, data=datadir, flags=flags)

        _run(cwd, "build/concorde")
        _run("make", "build/concorde")

        shutil.copyfile("build/concorde/concorde.a", "data/concorde.a")
        shutil.copyfile("build/concorde/concorde.h", "data/concorde.h")


class build_ext(_build_ext, object):
    """Build command that downloads and installs Concorde, if not found."""

    def run(self):
        if not self.has_external_concorde:
            download_concorde_qsopt()
            build_concorde()
        else:
            print("Using external Concorde/QSOpt")

        super(build_ext, self).run()

    @property
    def has_external_concorde(self):
        qsopt_dir = os.environ.get("QSOPT_DIR")
        concorde_dir = os.environ.get("CONCORDE_DIR")
        return bool(qsopt_dir) and bool(concorde_dir)


class ConcordeExtension(Extension, object):
    """Extension that sets Concorde/QSOpt lib/include args."""

    def __init__(self, *args, **kwargs):
        super(ConcordeExtension, self).__init__(*args, **kwargs)
        qsopt_dir = os.environ.get("QSOPT_DIR", "data")
        concorde_dir = os.environ.get("CONCORDE_DIR", "data")
        self.include_dirs.append(concorde_dir)
        self.extra_objects.extend(
            [pjoin(qsopt_dir, "qsopt.a"), pjoin(concorde_dir, "concorde.a")]
        )


setup(
    name="pyconcorde",
    ext_modules=cythonize(
        [
            ConcordeExtension(
                "concorde._concorde",
                sources=["concorde/_concorde.pyx"],
                include_dirs=[np.get_include()],
            )
        ]
    ),
    version="0.1.0",
    install_requires=[
        "cython>=0.22.0",
        "numpy>=1.21.0",
        "tsplib95",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    license="BSD",
    author="Joris Vankerschaver",
    author_email="joris.vankerschaver@gmail.com",
    url="https://github.com/jvkersch/pyconcorde",
    description="Cython wrappers for the Concorde TSP library",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
    ],
    cmdclass={
        "build_ext": build_ext,
    },
)



import math
import tsplib95 as tsplib

from tsplib95 import transformers, distances


distances.TYPES.update(
    {
        "EXACT_2D": math.dist,
        "EXACT_3D": math.dist,
        "FLOOR_2D": distances.euclidean,
        "FLOOR_3D": distances.euclidean,
    }
)


class NodeListField(tsplib.fields.TransformerField):
    default = list

    @classmethod
    def build_transformer(cls):
        node = transformers.FuncT(func=int)
        return transformers.ListT(value=node, terminal="-1")


class CTSPSetField(tsplib.fields.TransformerField):
    default = dict

    @classmethod
    def build_transformer(cls):
        salesman = transformers.FuncT(func=int)
        nodes = transformers.ListT(value=transformers.FuncT(func=int), terminal="-1")
        return transformers.MapT(key=salesman, value=nodes, sep="\n")


# demand can be multi-dimensional, for example in mPDTSP
class DemandsField(tsplib.fields.TransformerField):
    default = dict

    @classmethod
    def build_transformer(cls):
        node = transformers.FuncT(func=int)
        demand = transformers.ListT(value=transformers.FuncT(func=int))
        return transformers.MapT(key=node, value=demand, sep="\n")

    # we can only validate that the demands have the same dimensionality, not that it matches DEMAND_DIMENSION
    def validate(self, value):
        super().validate(value)
        dimensions = set(len(demand) for demand in value.values())
        if len(dimensions) > 1:
            error = (
                "all demands must have the same dimensionality "
                f"but got multiple dimensions {dimensions}"
            )
            raise tsplib.exceptions.ValidationError(error)


class LKHProblem(tsplib.models.StandardProblem):
    # extra spec fields
    demand_dimension = tsplib.fields.IntegerField("DEMAND_DIMENSION")
    distance = tsplib.fields.NumberField("DISTANCE")
    risk_threshold = tsplib.fields.IntegerField("RISK_THRESHOLD")
    salesmen = tsplib.fields.IntegerField("SALESMEN")
    scale = tsplib.fields.IntegerField("SCALE")
    service_time = tsplib.fields.NumberField("SERVICE_TIME")
    vehicles = tsplib.fields.IntegerField("VEHICLES")

    # extra data fields
    backhaul_section = NodeListField("BACKHAUL_SECTION")
    ctsp_set_section = CTSPSetField("CTSP_SET_SECTION")
    demand_section = DemandsField(
        "DEMAND_SECTION"
    )  # draft limit has same unit as demand
    draft_limit_section = DemandsField(
        "DRAFT_LIMIT_SECTION"
    )  # draft limit has same unit as demand
    pickup_and_delivery_section = tsplib.fields.MatrixField(
        "PICKUP_AND_DELIVERY_SECTION"
    )
    required_nodes_section = NodeListField("REQUIRED_NODES_SECTION")
    service_time_section = tsplib.fields.MatrixField("SERVICE_TIME_SECTION")
    time_window_section = tsplib.fields.MatrixField("TIME_WINDOW_SECTION")

    # need to override `render` because spec fields must precede data fields according to TSPLIB format
    # we assume that data fields end with _SECTION, and we sort those to the end of the field list
    # adapted from https://github.com/rhgrant10/tsplib95/blob/master/tsplib95/models.py#L217
    def render(self):
        # render each value by keyword
        rendered = self.as_name_dict()
        for name in list(rendered):
            value = rendered.pop(name)
            field = self.__class__.fields_by_name[name]
            if name in self.__dict__ or value != field.get_default_value():
                rendered[field.keyword] = field.render(value)

        # build keyword-value pairs with the separator
        kvpairs = []
        for keyword, value in sorted(
            rendered.items(), key=lambda item: item[0].endswith("_SECTION")
        ):
            sep = ":\n" if "\n" in value else ": "
            kvpairs.append(f"{keyword}{sep}{value}")
        kvpairs.append("EOF")

        # join and return the result
        return "\n".join(kvpairs)



import importlib.util

# base
from .data import TSPLIBOriDataset, TSPUniformDataset
from .data import SATLIBOriDataset
from .evaluate import TSPEvaluator, TSPLIBOriEvaluator, TSPUniformEvaluator
from .evaluate import SATLIBEvaluator
from .generator import TSPDataGenerator, MISDataGenerator
from .solver import TSPSolver, TSPLKHSolver, TSPConcordeSolver
from .solver import MISSolver, KaMISSolver, MISGurobi
from .utils import download, compress_folder, extract_archive, _get_md5

# expand
found_matplotlib = importlib.util.find_spec("matplotlib")
if found_matplotlib is not None:
    from .draw.tsp import draw_tsp_problem, draw_tsp_solution
    from .draw.mis import draw_mis_problem, draw_mis_solution
else:
    print("matplotlib not installed")


__version__ = "0.0.1a1"
__author__ = "SJTU-ReThinkLab"



import os
from ml4co_kit.utils import download, extract_archive


TSPLIB_LARGE_EUC_2D_PATH = "dataset/tsplib_original/large/EUC_2D"
TSPLIB_LARGE_EUC_2D = [
    "u1060",
    "vm1084",
    "pcb1173",
    "d1291",
    "rl1304",
    "rl1323",
    "nrw1379",
    "fl1400",
    "u1432",
    "fl1577",
    "d1655",
    "vm1748",
    "u1817",
    "rl1889",
    "d2103",
    "u2152",
    "u2319",
    "pcb3038",
    "fl3795",
    "fnl4461",
    "rl5915",
    "rl5934",
    "rl11849",
    "usa13509",
    "brd14051",
    "d15112",
    "d18512",
]

TSPLIB_LARGE_CEIL_2D_PATH = "dataset/tsplib_original/large/CEIL_2D"
TSPLIB_LARGE_CEIL_2D = ["pla7397", "pla33810", "pla85900"]

TSPLIB_RESOLVE_EUC_2D_PATH = "dataset/tsplib_original/resolved/EUC_2D/problem"
TSPLIB_RESOLVE_EUC_2D_SOLUTION = "dataset/tsplib_original/resolved/EUC_2D/solution"
TSPLIB_RESOLVE_EUC_2D = [
    "att48",
    "eil51",
    "berlin52",
    "st70",
    "eil76",
    "pr76",
    "kroA100",
    "kroC100",
    "kroD100",
    "rd100",
    "eil101",
    "lin105",
    "ch130",
    "ch150",
    "tsp225",
    "a280",
    "pr1002",
    "pr2392",
]

TSPLIB_RESOLVE_EXPLICIT_PATH = "dataset/tsplib_original/resolved/EXPLICIT/problem"
TSPLIB_RESOLVE_EXPLICIT_SOLUTION = "dataset/tsplib_original/resolved/EXPLICIT/solution"
TSPLIB_RESOLVE_EXPLICIT = [
    "gr24",
    "fri26",
    "bayg29",
    "bays29",
    "gr48",
    "gr120",
    "brg180",
    "pa561",
]

TSPLIB_RESOLVE_GEO_PATH = "dataset/tsplib_original/resolved/GEO/problem"
TSPLIB_RESOLVE_GEO_SOLUTION = "dataset/tsplib_original/resolved/GEO/solution"
TSPLIB_RESOLVE_GEO = ["ulysses16", "ulysses22", "gr96", "gr202", "gr666"]

TSPLIB_UNRESOLVE_EUC_2D_PATH = "dataset/tsplib_original/unresolved/EUC_2D"
TSPLIB_UNRESOLVE_EUC_2D = [
    "rat99",
    "kroB100",
    "kroE100",
    "pr107",
    "pr124",
    "bier127",
    "pr136",
    "pr144",
    "kroA150",
    "kroB150",
    "pr152",
    "u159",
    "rat195",
    "d198",
    "kroA200",
    "kroB200",
    "ts225",
    "pr226",
    "gil262",
    "pr264",
    "pr299",
    "lin318",
    "linhp318",
    "rd400",
    "fl417",
    "pr439",
    "pcb442",
    "d493",
    "att532",
    "u574",
    "rat575",
    "p654",
    "d657",
    "u724",
    "rat783",
    "dsj1000",
]

TSPLIB_UNRESOLVE_EXPLICIT_PATH = "dataset/tsplib_original/unresolved/EXPLICIT"
TSPLIB_UNRESOLVE_EXPLICIT = ["gr17", "gr21", "dantzig42", "swiss42", "hk48", "brazil58"]

TSPLIB_UNRESOLVE_GEO_PATH = "dataset/tsplib_original/unresolved/GEO"
TSPLIB_UNRESOLVE_GEO = ["burma14", "ali535"]


class TSPLIBOriDataset:
    def __init__(self):
        self.url = "https://huggingface.co/datasets/ML4CO/TSPLIBOriDataset/resolve/main/tsplib_original.tar.gz?download=true"
        self.md5 = "1dc8dc14888808bb29167ac4fe052983"
        self.dir = "dataset/tsplib_original"
        self.raw_data_path = "dataset/tsplib_original.tar.gz"
        if not os.path.exists("dataset"):
            os.mkdir("dataset")
        if not os.path.exists(self.dir):
            download(filename=self.raw_data_path, url=self.url, md5=self.md5)
            extract_archive(archive_path=self.raw_data_path, extract_path=self.dir)

    @property
    def support(self):
        return {
            "large": {
                "EUC_2D": {
                    "path": TSPLIB_LARGE_EUC_2D_PATH,
                    "problem": TSPLIB_LARGE_EUC_2D,
                },
                "CEIL_2D": {
                    "path": TSPLIB_LARGE_CEIL_2D_PATH,
                    "problem": TSPLIB_LARGE_CEIL_2D,
                },
            },
            "resolved": {
                "EUC_2D": {
                    "path": TSPLIB_RESOLVE_EUC_2D_PATH,
                    "solution": TSPLIB_RESOLVE_EUC_2D_SOLUTION,
                    "problem": TSPLIB_RESOLVE_EUC_2D,
                },
                "EXPLICIT": {
                    "path": TSPLIB_RESOLVE_EXPLICIT_PATH,
                    "solution": TSPLIB_RESOLVE_EXPLICIT_SOLUTION,
                    "problem": TSPLIB_RESOLVE_EXPLICIT,
                },
                "GEO": {
                    "path": TSPLIB_RESOLVE_GEO_PATH,
                    "solution": TSPLIB_RESOLVE_GEO_SOLUTION,
                    "problem": TSPLIB_RESOLVE_GEO,
                },
            },
            "unresolved": {
                "EUC_2D": {
                    "path": TSPLIB_UNRESOLVE_EUC_2D_PATH,
                    "problem": TSPLIB_UNRESOLVE_EUC_2D,
                },
                "EXPLICIT": {
                    "path": TSPLIB_UNRESOLVE_EXPLICIT_PATH,
                    "problem": TSPLIB_UNRESOLVE_EXPLICIT,
                },
                "GEO": {
                    "path": TSPLIB_UNRESOLVE_GEO_PATH,
                    "problem": TSPLIB_UNRESOLVE_GEO,
                },
            },
        }



import os
import sys

root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root_folder)
from ml4co_kit.solver import TSPSolver, TSPLKHSolver, TSPConcordeSolver, KaMISSolver
from ml4co_kit.utils.mis_utils import cnf_folder_to_gpickle_folder


##############################################
#             Test Func For TSP              #
##############################################

def test_tsp_base_solver():
    solver = TSPSolver()
    solver.from_txt("tests/solver_test/tsp50_test_small.txt")
    os.remove("tests/solver_test/tsp50_test_small.txt")
    solver.read_tours(solver.ref_tours)
    solver.to_tsp(
        save_dir="tests/solver_test/tsp50_test_small",
        filename="problem"
    )
    solver.to_opt_tour(
        save_dir="tests/solver_test/tsp50_test_small",
        filename="solution"
    )
    solver.to_txt("tests/solver_test/tsp50_test_small.txt")
    

def _test_tsp_lkh_solver(show_time: bool, num_threads: int):
    tsp_lkh_solver = TSPLKHSolver(lkh_max_trials=100)
    tsp_lkh_solver.from_txt("tests/solver_test/tsp50_test.txt")
    tsp_lkh_solver.solve(show_time=show_time, num_threads=num_threads)
    _, _, gap_avg, _ = tsp_lkh_solver.evaluate(calculate_gap=True)
    print(f"TSPLKHSolver Gap: {gap_avg}")
    if gap_avg >= 1e-2:
        message = (
            f"The average gap ({gap_avg}) of TSP50 solved by TSPLKHSolver "
            "is larger than or equal to 1e-2%."
        )
        raise ValueError(message)


def test_tsp_lkh_solver():
    _test_tsp_lkh_solver(True, 1)
    _test_tsp_lkh_solver(True, 2)
    _test_tsp_lkh_solver(False, 1)
    _test_tsp_lkh_solver(False, 2)
    

def _test_tsp_concorde_solver(show_time: bool, num_threads: int):
    tsp_lkh_solver = TSPConcordeSolver()
    tsp_lkh_solver.from_txt("tests/solver_test/tsp50_test.txt")
    tsp_lkh_solver.solve(show_time=show_time, num_threads=num_threads)
    _, _, gap_avg, _ = tsp_lkh_solver.evaluate(calculate_gap=True)
    print(f"TSPConcordeSolver Gap: {gap_avg}")
    if gap_avg >= 1e-3:
        message = (
            f"The average gap ({gap_avg}) of TSP50 solved by TSPConcordeSolver "
            "is larger than or equal to 1e-3%."
        )
        raise ValueError(message)


def test_tsp_concorde_solver():
    _test_tsp_concorde_solver(True, 1)
    _test_tsp_concorde_solver(True, 2)
    _test_tsp_concorde_solver(False, 1)
    _test_tsp_concorde_solver(False, 2)


def test_tsp():
    """
    Test TSPSolver
    """
    test_tsp_base_solver()
    test_tsp_lkh_solver()
    test_tsp_concorde_solver()


##############################################
#            Test Func For KaMIS             #
##############################################


def _test_kamis_solver():
    kamis_solver = KaMISSolver(time_limit=30)
    cnf_folder_to_gpickle_folder(
        cnf_folder="tests/solver_test/mis_test_cnf",
        gpickle_foler="tests/solver_test/mis_test"
    )
    kamis_solver.solve(
        src="tests/solver_test/mis_test/mis_graph", 
        out="tests/solver_test/mis_test/mis_graph/solve"
    )
    kamis_solver.from_gpickle_folder("tests/solver_test/mis_test/mis_graph")
    kamis_solver.read_ref_sel_nodes_num_from_txt("tests/solver_test/mis_test/ref_solution.txt")
    gap_avg = kamis_solver.evaluate(calculate_gap=True)["avg_gap"]
    print(f"KaMISSolver Gap: {gap_avg}")
    if gap_avg >= 0.1:
        message = (
            f"The average gap ({gap_avg}) of MIS solved by KaMISSolver "
            "is larger than or equal to 0.1%."
        )
        raise ValueError(message)


def test_mis():
    """
    Test MISSolver
    """
    _test_kamis_solver()


##############################################
#                    MAIN                    #
##############################################

if __name__ == "__main__":
    test_tsp()
    test_mis()



import os
import time
import requests
import shutil
import aiohttp
import asyncio
import hashlib
import zipfile
import tarfile
import async_timeout
import urllib.request
from tqdm import tqdm
from typing import Union


###############################################
#                  Download                   #
###############################################


def download(filename: str, url: Union[str, list], md5: str = None, retries: int = 5):
    if type(url) == str:
        return _download(filename, url, md5, retries)
    elif type(url) == list:
        for cur_url in url:
            try:
                return _download(filename, cur_url, md5, retries)
            except RuntimeError:
                continue
        raise RuntimeError("Max Retries exceeded!")
    else:
        raise ValueError("The url should be string or list of string.")


async def _asyncdownload(filename: str, url: str):
    async with aiohttp.ClientSession() as session:
        async with async_timeout.timeout(120):
            async with session.get(url) as response:
                with open(filename, "wb") as file:
                    async for data in response.content.iter_chunked(512):
                        file.write(data)


def _download(filename: str, url: str, md5: str, retries: int):
    if retries <= 0:
        raise RuntimeError("Max Retries exceeded!")
    if not os.path.exists(filename):
        print(f"\nDownloading to {filename}...")
        if retries % 3 == 1:
            try:
                down_res = requests.get(url, stream=True)
                file_size = int(down_res.headers.get("Content-Length", 0))
                with tqdm.wrapattr(down_res.raw, "read", total=file_size) as content:
                    with open(filename, "wb") as file:
                        shutil.copyfileobj(content, file)
            except requests.exceptions.ConnectionError as err:
                print("Warning: Network error. Retrying...\n", err)
                return download(filename, url, md5, retries - 1)
        elif retries % 3 == 2:
            try:
                asyncio.run(_asyncdownload(filename, url))
            except:
                return _download(filename, url, md5, retries - 1)
        else:
            try:
                urllib.request.urlretrieve(url, filename)
            except:
                return _download(filename, url, md5, retries - 1)

    if md5 is not None:
        md5_returned = _get_md5(filename)
        if md5 != md5_returned:
            print("Warning: MD5 check failed for the downloaded content. Retrying...")
            os.remove(filename)
            time.sleep(1)
            return _download(filename, url, md5, retries - 1)
    return filename


def _get_md5(filename: str):
    hash_md5 = hashlib.md5()
    chunk = 8192
    with open(filename, "rb") as file_to_check:
        while True:
            buffer = file_to_check.read(chunk)
            if not buffer:
                break
            hash_md5.update(buffer)
        md5_returned = hash_md5.hexdigest()
        return md5_returned


###############################################
#                  Compress                   #
###############################################


def compress_folder(
    folder: str,
    compress_path: str,
):
    """
    Compresses a folder into the specified output format.

    Args:
        folder (str):
            Path to the folder to be compressed.
        compress_path (str):
            Output file path.

    Supported formats:
        - .zip: ZIP format
        - .tar.gz: tar.gz format
    """
    if compress_path.endswith(".zip"):
        shutil.make_archive(compress_path[:-4], "zip", folder)
    elif compress_path.endswith(".tar.gz"):
        with tarfile.open(compress_path, "w:gz") as tar:
            tar.add(folder, arcname="")
    else:
        message = "Unsupported file format. Only .zip, .tar.gz"
        raise ValueError(message)


def extract_archive(archive_path: str, extract_path: str):
    """
    Extracts an archive into the specified extract path.

    Args:
        archive_path (str):
            Path to the archive file.
        extract_path (str):
            Path to the extraction directory.

    Supported formats:
        - .zip: ZIP format
        - .tar.gz: tar.gz format
    """
    if archive_path.endswith(".zip"):
        with zipfile.ZipFile(archive_path, "r") as zip_ref:
            zip_ref.extractall(extract_path)
    elif archive_path.endswith(".tar.gz"):
        with tarfile.open(archive_path, "r:gz") as tar_ref:
            tar_ref.extractall(extract_path)
    else:
        message = "Unsupported file format. Only .zip, .tar.gz"
        raise ValueError(message)



import time
import numpy as np
import tsplib95
import pathlib
from tqdm import tqdm
from multiprocessing import Pool
from typing import Union
from .base import TSPSolver
from .lkh_solver import lkh_solve


class TSPLKHSolver(TSPSolver):
    def __init__(
        self,
        lkh_max_trials: int = 1000,
        lkh_path: pathlib.Path = "LKH",
        scale: int = 1e6,
        lkh_runs: int = 10,
    ):
        """
        TSPLKHSolver
        Args:
            lkh_max_trials (int, optional): The maximum number of trials for
                the LKH solver. Defaults to 1000.
            lkh_path (pathlib.Path, optional): The path to the LKH solver.
                Defaults to "LKH".
            scale (int, optional): The scale factor for coordinates in the
                LKH solver. Defaults to 1e6.
            lkh_runs (int, optional): The number of runs for the LKH solver.
                Defaults to 10.
        """
        super(TSPLKHSolver, self).__init__(solver_type="lkh", scale=scale)
        self.lkh_max_trials = lkh_max_trials
        self.lkh_path = lkh_path
        self.scale = scale
        self.lkh_runs = lkh_runs

    def _solve(self, nodes_coord: np.ndarray) -> list:
        problem = tsplib95.models.StandardProblem()
        problem.name = "TSP"
        problem.type = "TSP"
        problem.dimension = self.nodes_num
        problem.edge_weight_type = self.norm
        problem.node_coords = {
            n + 1: nodes_coord[n] * self.scale for n in range(self.nodes_num)
        }
        solution = lkh_solve(
            solver=self.lkh_path,
            problem=problem,
            max_trials=self.lkh_max_trials,
            runs=self.lkh_runs,
        )
        tour = [n - 1 for n in solution[0]]
        return tour

    def solve(
        self,
        points: Union[np.ndarray, list] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        num_threads: int = 1,
        show_time: bool = False,
    ) -> np.ndarray:
        # prepare
        self.from_data(points, norm, normalize)
        start_time = time.time()

        # solve
        tours = list()
        p_shape = self.points.shape
        num_points = p_shape[0]
        if num_threads == 1:
            if show_time:
                for idx in tqdm(range(num_points), desc="Solving TSP Using LKH"):
                    tours.append(self._solve(self.points[idx]))
            else:
                for idx in range(num_points):
                    tours.append(self._solve(self.points[idx]))
        else:
            batch_points = self.points.reshape(
                -1, num_threads, p_shape[-2], p_shape[-1]
            )
            if show_time:
                for idx in tqdm(
                    range(num_points // num_threads), desc="Solving TSP Using LKH"
                ):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.map(
                            self._solve,
                            [
                                batch_points[idx][inner_idx]
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)
            else:
                for idx in range(num_points // num_threads):
                    with Pool(num_threads) as p1:
                        cur_tours = p1.map(
                            self._solve,
                            [
                                batch_points[idx][inner_idx]
                                for inner_idx in range(num_threads)
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)

        # format
        tours = np.array(tours)
        zeros = np.zeros((tours.shape[0], 1))
        tours = np.append(tours, zeros, axis=1).astype(np.int32)
        if tours.ndim == 2 and tours.shape[0] == 1:
            tours = tours[0]
        self.read_tours(tours)
        end_time = time.time()
        if show_time:
            print(f"Use Time: {end_time - start_time}")
        return tours

    def regret_solve(
        self, points: np.ndarray, fixed_edges: tuple, norm: str = "EUC_2D"
    ):
        problem = tsplib95.models.StandardProblem()
        problem.name = "TSP"
        problem.type = "TSP"
        problem.dimension = points.shape[0]
        problem.edge_weight_type = norm
        problem.node_coords = {
            n + 1: self.scale * points[n] for n in range(points.shape[0])
        }
        problem.fixed_edges = [[n + 1 for n in fixed_edges]]
        solution = lkh_solve(
            solver=self.lkh_path,
            problem=problem,
            max_trials=self.lkh_max_trials,
            runs=self.lkh_runs,
        )
        tour = [n - 1 for n in solution[0]] + [0]
        np_tour = np.array(tour)
        return np_tour



import math
import numpy as np
import scipy.spatial
from typing import Union

SUPPORT_TSPLIB_TYPE = ["EUC_2D", "GEO"]


class TSPEvaluator(object):
    def __init__(self, points: Union[list, np.ndarray], norm: str = "EUC_2D"):
        if type(points) == list:
            points = np.array(points)
        if points.ndim == 3 and points.shape[0] == 1:
            points = points[0]
        if points.ndim != 2:
            raise ValueError("points must be 2D array.")
        self.points = points
        self.set_norm(norm)
        self.edge_weights = self.get_weight()

    def set_norm(self, norm: str):
        if norm not in SUPPORT_TSPLIB_TYPE:
            message = (
                f"The norm type ({norm}) is not a valid type, "
                f"only {SUPPORT_TSPLIB_TYPE} are supported."
            )
            raise ValueError(message)
        self.norm = norm

    def get_weight(self):
        if self.norm == "EUC_2D":
            return scipy.spatial.distance_matrix(self.points, self.points)
        elif self.norm == "GEO":
            nodes_num = self.points.shape[0]
            edge_weights = np.zeros(shape=(nodes_num, nodes_num), dtype=np.int32)
            for i_idx in range(nodes_num):
                for j_idx in range(nodes_num):
                    if j_idx <= i_idx:
                        continue
                    node_coord_1 = self.points[i_idx]
                    node_coord_2 = self.points[j_idx]
                    weight = geographical(node_coord_1, node_coord_2)
                    edge_weights[i_idx][j_idx] = weight
                    edge_weights[j_idx][i_idx] = weight
            return edge_weights

    def evaluate(self, route):
        total_cost = 0
        for i in range(len(route) - 1):
            total_cost += self.edge_weights[route[i], route[i + 1]]
        return total_cost


def parse_degrees(coord):
    """Parse an encoded geocoordinate value into real degrees.

    :param float coord: encoded geocoordinate value
    :return: real degrees
    :rtype: float
    """
    degrees = int(coord)
    minutes = coord - degrees
    return degrees + minutes * 5 / 3


class RadianGeo:
    def __init__(self, coord):
        x, y = coord
        self.lat = self.__class__.parse_component(x)
        self.lng = self.__class__.parse_component(y)

    @staticmethod
    def parse_component(component):
        return math.radians(parse_degrees(component))


def geographical(start, end, radius=6378.388):
    """Return the geographical distance between start and end.

    This is capable of performing distance calculations for GEO problems.

    :param tuple start: *n*-dimensional coordinate
    :param tuple end: *n*-dimensional coordinate
    :param float radius: the radius of the Earth
    :return: rounded distance
    """
    if len(start) != len(end):
        raise ValueError("dimension mismatch between start and end")

    start = RadianGeo(start)
    end = RadianGeo(end)

    q1 = math.cos(start.lng - end.lng)
    q2 = math.cos(start.lat - end.lat)
    q3 = math.cos(start.lat + end.lat)
    distance = radius * math.acos(0.5 * ((1 + q1) * q2 - (1 - q1) * q3)) + 1

    return distance



import io
import re
import os
import sys
import time
import json
import copy
import pathlib
import pulp as plp
import numpy as np
import networkx as nx
import pandas as pd
from pathlib import Path
from .base import MISSolver


class MISGurobi(MISSolver):
    def __init__(
        self,
        weighted: bool = False,
        time_limit: float = 60.0,
        num_threads: int = 8,
        quadratic: bool = False,
        write_mps: bool = False,
    ):
        """
        MISGurobi
        Args:
            weighted (bool, optional):
                If enabled, solve the weighted MIS problem instead of MIS.
            time_limit (float, optional):
                Time limit in seconds.
            num_threads (int, optional):
                Maximum number of threads to use.
            quadratic (bool, optional):
                Whether a quadratic program should be used instead of a linear program
                to solve the MIS problem (cannot be used together with weighted).
            write_mps (bool, optional):
                Instead of solving, write mps output (e.g., for tuning)
        """
        super(MISGurobi, self).__init__()
        self.solver_type = "gurobi"
        self.weighted = weighted
        self.time_limit = time_limit
        self.num_threads = num_threads
        self.quadratic = quadratic
        self.write_mps = write_mps
        self.gurobi_path = pathlib.Path(__file__).parent

    @staticmethod
    def __prepare_graph(g: nx.Graph, weighted=False):
        graph = copy.deepcopy(g)
        graph.remove_edges_from(nx.selfloop_edges(graph))
        # the gurobi solver file always expects a weighted file
        # however, for "unweighted" we supply all weights as 1
        if not weighted:
            nx.set_node_attributes(graph, 1, name="weight")
        return graph

    def prepare_instances(
        self, instance_directory: pathlib.Path, cache_directory: pathlib.Path
    ):
        for graph_path in instance_directory.rglob("*.gpickle"):
            self.prepare_instance(graph_path.resolve(), cache_directory)

    def prepare_instance(
        self,
        source_instance_file: pathlib.Path,
        cache_directory: pathlib.Path,
    ):
        source_instance_file = Path(source_instance_file)
        cache_directory = Path(cache_directory)
        cache_directory.mkdir(parents=True, exist_ok=True)

        dest_path = cache_directory / (
            source_instance_file.stem
            + f"_{'weighted' if self.weighted else 'unweighted'}.graph"
        )

        if os.path.exists(dest_path):
            source_mtime = os.path.getmtime(source_instance_file)
            last_updated = os.path.getmtime(dest_path)
            if source_mtime <= last_updated:
                return  # we already have an up2date version of that file

        print(f"Updated graph file: {source_instance_file}.")
        g = nx.read_gpickle(source_instance_file)
        graph = self.__prepare_graph(g, weighted=self.weighted)
        nx.write_gpickle(graph, dest_path)

    def solve(self, src: pathlib.Path, out: pathlib.Path):
        print("Solving all given instances using " + str(self))
        # preprocess
        src = Path(src)
        out = Path(out)
        cache_directory = src / "preprocessed"
        self.prepare_instances(src, cache_directory)
        # call gurobi
        results = {}
        for _, graph_path in enumerate(
            cache_directory.rglob(
                f"*_{'weighted' if self.weighted else 'unweighted'}.graph"
            )
        ):
            graph_name_stub = os.path.splitext(os.path.basename(graph_path))[0].rsplit(
                "_", 1
            )[0]
            print(f"Solving graph {graph_path}")
            graph = nx.read_gpickle(graph_path)
            if self.write_mps:
                self._solve(graph, graph_name_stub, out)
            else:
                solu, sum, status, total_time, explore_time = self._solve(
                    graph, graph_name_stub, out
                )
                print(f"Found MWIS: n={len(solu)}, w={sum} ✔️✔️")
                if status != "Optimal":
                    print(f"Non-Optimal Gurobi status: {status}")
                results[graph_name_stub] = {
                    "total_time": total_time,
                    "mwis_vertices": int(len(solu)),
                    "mwis_weight": float(sum),
                    "mwis": np.ravel(solu).tolist(),
                    "gurobi_status": status,
                    "gurobi_explore_time": explore_time,
                }
        # write the result
        if not self.write_mps:
            with open(out / "results.json", "w", encoding="utf-8") as f:
                json.dump(results, f, ensure_ascii=False, sort_keys=True, indent=4)
        print("Done with all graphs, exiting.")

    def _solve(self, graph: nx.Graph, graph_name_stub: str, out: pathlib.Path):
        start_time = time.monotonic()
        weight_dict = nx.get_node_attributes(graph, "weight")
        # not defined whether dictionary returned by networkx is sorted
        weight_list = [weight_dict[key] for key in sorted(weight_dict.keys())]
        wts = np.array(weight_list)

        if not self.quadratic:
            self.non_quadratic_solve(
                start_time=start_time,
                graph=graph,
                wts=wts,
                graph_name_stub=graph_name_stub,
                out=out,
            )
        else:
            self.quadratic_solve(
                start_time=start_time,
                graph=graph,
                wts=wts,
                graph_name_stub=graph_name_stub,
                out=out,
            )

    def quadratic_solve(
        self,
        start_time: float,
        graph: nx.Graph,
        wts: np.ndarray,
        graph_name_stub: str,
        out: pathlib.Path,
    ):
        import gurobipy as gp
        from gurobipy import GRB

        n = graph.number_of_nodes()
        adj = nx.to_numpy_array(graph)
        J = np.identity(n)
        A = J - adj
        m = gp.Model("mis")
        x = m.addMVar(shape=n, vtype=GRB.BINARY, name="x")
        m.setObjective(x @ A @ x, GRB.MAXIMIZE)
        m.setParam("TimeLimit", self.time_limit)
        m.setParam("Threads", self.num_threads)
        m.setParam("ImproveStartTime", self.time_limit * 0.9)

        if self.write_mps:
            m.write(out / f"{graph_name_stub}.mps")
            return

        # redirect stdout by gurobi into internal string buffer
        real_stdout = sys.stdout
        sys.stdout = io.StringIO()
        m.optimize()
        gurobi_output = sys.stdout.getvalue()
        sys.stdout = real_stdout  # set stdout file back to normal
        print(gurobi_output)
        # parse solving time from gurobi output
        time_string_match = re.compile("in (\d+(\.\d*)) seconds").findall(gurobi_output)[0]
        apparent_solve_time = float(time_string_match[0])
        solu = np.nonzero(np.array(x.X))[0]

        # status
        status = m.status
        if status == GRB.OPTIMAL:
            status = "Optimal"
        else:
            status = str(status)
        end_time = time.monotonic()
        return solu, wts[solu].sum(), status, end_time - start_time, apparent_solve_time

    def non_quadratic_solve(
        self,
        start_time: float,
        graph: nx.Graph,
        wts: np.ndarray,
        graph_name_stub: str,
        out: pathlib.Path,
    ):
        opt_model = plp.LpProblem(name="model")
        adj = nx.adjacency_matrix(graph)
        x_vars = {
            i: plp.LpVariable(cat=plp.LpBinary, name="x_{0}".format(i))
            for i in range(wts.size)
        }
        # not sure why we need set here, as the call to range produces a unique list
        set_V = set(range(wts.size))
        constraints = dict()
        ei = 0  # number of constraints counter

        # we iterate over all nodes (could we be more explicit and use G.nodes here?)
        for j in set_V:
            # set N are all the vertices that j has a connection to
            _, set_N = np.nonzero(adj[j])
            for i in set_N:
                constraints[ei] = opt_model.addConstraint(
                    plp.LpConstraint(
                        e=plp.lpSum([x_vars[i], x_vars[j]]),
                        sense=plp.LpConstraintLE,
                        rhs=1,
                        # constraint that we only pick one of two neighboring nodes
                        name="constraint_{0}_{1}".format(j, i),
                    )
                )
                ei = ei + 1

        # we want to maximize the weight
        objective = plp.lpSum(x_vars[i] * wts[i] for i in set_V)
        opt_model.sense = plp.LpMaximize
        opt_model.setObjective(objective)

        if self.write_mps:
            opt_model.writeMPS(out / (graph_name_stub + ".mps"))
            return

        # redirect stdout by gurobi into internal string buffer
        real_stdout = sys.stdout
        sys.stdout = io.StringIO()

        # solve
        opt_model.solve(
            solver=plp.apis.GUROBI(
                mip=True,
                timeLimit=self.time_limit,
                Threads=self.num_threads,
                ImproveStartTime=self.time_limit * 0.9,
            )
        )

        gurobi_output = sys.stdout.getvalue()
        sys.stdout = real_stdout  # set stdout file back to normal
        print(gurobi_output)

        # parse solving time from gurobi output
        time_string_match = re.compile("in (\d+(\.\d*)) seconds").findall(
            gurobi_output
        )[0]
        apparent_solve_time = float(time_string_match[0])

        if plp.LpStatus[opt_model.status] == "Optimal":
            opt_df = pd.DataFrame.from_dict(
                x_vars, orient="index", columns=["variable_object"]
            )
            opt_df["solution_value"] = opt_df["variable_object"].apply(
                lambda item: item.varValue
            )
            solu = opt_df[opt_df["solution_value"] > 0].index.to_numpy()
        else:
            # Retrieve best (not optimal) solution
            # GUROBI specific, seems to be a bug in PuLP where varValue is None for non-optimal solutions
            solu = np.nonzero(np.array(opt_model.solverModel.X))[0]

        end_time = time.monotonic()
        return (
            solu,
            wts[solu].sum(),
            plp.LpStatus[opt_model.status],
            end_time - start_time,
            apparent_solve_time,
        )

    def __str__(self) -> str:
        return "gurobi"



# -*- coding: utf-8 -*-
from __future__ import division, print_function

from collections import namedtuple
import os
import shutil
import tempfile
import uuid
import numpy as np

from ._concorde import _CCutil_gettsplib, _CCtsp_solve_dat

ComputedTour = namedtuple(
    "ComputedTour", ["tour", "optimal_value", "success", "found_tour", "hit_timebound"]
)


EDGE_WEIGHT_TYPES = {
    "EXPLICIT",
    "EUC_2D",
    "EUC_3D",
    "MAX_2D",
    "MAN_2D",
    "GEO",
    "GEOM",
    "ATT",
    "CEIL_2D",
    "DSJRAND",
}


def write_tsp_file(fp, xs, ys, norm, name):
    """Write data to a TSPLIB file."""
    if len(xs) != len(ys):
        raise ValueError(
            "x and y coordinate vector must have the "
            "same length ({} != {})".format(len(xs), len(ys))
        )
    if norm not in EDGE_WEIGHT_TYPES:
        raise ValueError(
            "Norm {!r} must be one of {}".format(norm, ", ".join(EDGE_WEIGHT_TYPES))
        )

    fp.write("NAME: {}\n".format(name))
    fp.write("TYPE: TSP\n")
    fp.write("DIMENSION: {}\n".format(len(xs)))
    fp.write("EDGE_WEIGHT_TYPE: {}\n".format(norm))
    fp.write("NODE_COORD_SECTION\n")
    for n, (x, y) in enumerate(zip(xs, ys), start=1):
        fp.write("{} {} {}\n".format(n, x, y))
    fp.write("EOF\n")




class TSPSolver(object):
    def __init__(self):
        self._data = None
        self._ncount = -1

    @classmethod
    def from_tspfile(cls, fname):
        ncount, data = _CCutil_gettsplib(fname)
        if data is None:
            raise RuntimeError("Error in loading {}".format(fname))
        self = cls()
        self._ncount = ncount
        self._data = data
        return self

    @classmethod
    def from_data(cls, xs, ys, norm, name=None):
        """Construct datagroup from given data.

        This routine writes the given data to a temporary file, and then uses
        Concorde's file parser to read from file and do the initialization.
        """
        if norm not in EDGE_WEIGHT_TYPES:
            raise ValueError(
                "norm must be one of {} but got {!r}".format(
                    ", ".join(EDGE_WEIGHT_TYPES), norm
                )
            )

        # TODO: properly figure out Concorde's CCdatagroup format and
        # initialize this object directly instead of going via file.
        if name is None:
            name = uuid.uuid4().hex
        try:
            ccdir = tempfile.mkdtemp()
            ccfile = os.path.join(ccdir, "data.tsp")
            with open(ccfile, "w") as fp:
                write_tsp_file(fp, xs, ys, norm, name)
            return cls.from_tspfile(ccfile)
        finally:
            shutil.rmtree(ccdir)

    @property
    def x(self):
        return self._data.x

    @property
    def y(self):
        return self._data.y

    @property
    def z(self):
        return self._data.z

    def __str__(self):
        if self._data is None:
            return "Uninitialized TSPSolver"
        else:
            return "TSPSolver with {} nodes".format(self._ncount)

    def solve(
        self, 
        name: str, 
        time_bound: int=-1, 
        verbose: bool=True, 
        random_seed: int=0
    ):
        name = name[0:9]
        res = _CCtsp_solve_dat(
            self._ncount, self._data, name, time_bound, not verbose, random_seed
        )
        return ComputedTour(*res)



import os
import tsplib95
import numpy as np
from typing import Union
from ml4co_kit.evaluate.tsp.base import TSPEvaluator


SUPPORT_TSPLIB_TYPE = ["EUC_2D", "GEO"]


class TSPSolver:
    def __init__(self, solver_type: str = None, scale: int = 1e6):
        self.solver_type = solver_type
        self.scale = scale
        self.points = None
        self.norm = None
        self.ori_points = None
        self.tours = None
        self.ref_tours = None
        self.nodes_num = None

    def check_points_dim(self):
        self.check_ori_points_dim()
        if self.points is None:
            return
        elif self.points.ndim == 2:
            self.points = np.expand_dims(self.points, axis=0)
        if self.points.ndim != 3:
            raise ValueError("``points`` must be a 2D or 3D array.")
        self.nodes_num = self.points.shape[1]

    def check_ori_points_dim(self):
        if self.ori_points is None:
            return
        elif self.ori_points.ndim == 2:
            self.ori_points = np.expand_dims(self.ori_points, axis=0)
        if self.ori_points.ndim != 3:
            raise ValueError("The ``ori_points`` must be 2D or 3D array.")

    def check_tours_dim(self):
        if self.tours is None:
            return
        elif self.tours.ndim == 1:
            self.tours = np.expand_dims(self.tours, axis=0)
        if self.tours.ndim != 2:
            raise ValueError("The dimensions of ``tours`` cannot be larger than 2.")

    def check_ref_tours_dim(self):
        if self.ref_tours is None:
            return
        elif self.ref_tours.ndim == 1:
            self.ref_tours = np.expand_dims(self.ref_tours, axis=0)
        if self.ref_tours.ndim != 2:
            raise ValueError(
                "The dimensions of the ``ref_tours`` cannot be larger than 2."
            )

    def check_points_not_none(self):
        if self.points is None:
            message = (
                "``points`` cannot be None! You can load the points using the methods"
                "``from_data``, ``from_txt``, or ``from_tsp``."
            )
            raise ValueError(message)

    def check_tours_not_none(self):
        if self.tours is None:
            message = (
                "``tours`` cannot be None! You can use solvers based on ``TSPSolver`` "
                "like ``TSPLKHSolver`` or the method ``read_tours`` to get the tours."
            )
            raise ValueError(message)

    def check_ref_tours_not_none(self):
        if self.ref_tours is None:
            message = (
                "``ref_tours`` cannot be None! You can use solvers based on ``TSPSolver`` "
                "like ``TSPLKHSolver`` to obtain the tours and set them as the ``ref_tours``. "
                "Or you can use the methods ``read_ref_tours``, ``from_txt``, "
                "or ``read_ref_tours_from_opt_tour`` to load them."
            )
            raise ValueError(message)

    def set_norm(self, norm: str):
        if norm is None:
            return
        if norm not in SUPPORT_TSPLIB_TYPE:
            message = (
                f"The norm type ({norm}) is not a valid type, "
                f"only {SUPPORT_TSPLIB_TYPE} are supported."
            )
            raise ValueError(message)
        if norm == "GEO" and self.scale != 1:
            message = "The scale must be 1 for ``GEO`` norm type."
            raise ValueError(message)
        self.norm = norm

    def normalize_points(self):
        for idx in range(self.points.shape[0]):
            cur_points = self.points[idx]
            max_value = np.max(cur_points)
            min_value = np.min(cur_points)
            cur_points = (cur_points - min_value) / (max_value - min_value)
            self.points[idx] = cur_points

    def from_tsp(self, file_path: str, norm: str = "EUC_2D", normalize: bool = False):
        self.set_norm(norm)
        if not file_path.endswith(".tsp"):
            raise ValueError("Invalid file format. Expected a ``.tsp`` file.")
        tsplib_data = tsplib95.load(file_path)
        points = np.array(list(tsplib_data.node_coords.values()))
        if points is None:
            raise RuntimeError("Error in loading {}".format(file_path))
        self.ori_points = points
        self.points = points.astype(np.float32)
        self.check_points_dim()
        if normalize:
            self.normalize_points()

    def from_txt(
        self,
        file_path: str,
        norm: str = "EUC_2D",
        normalize: bool = False,
        load_ref_tours: str = True,
        return_list: bool = False,
    ):
        self.set_norm(norm)

        # check the file format
        if not file_path.endswith(".txt"):
            raise ValueError("Invalid file format. Expected a ``.txt`` file.")

        # read the data form .txt
        with open(file_path, "r") as file:
            nodes_coords = list()
            ref_tours = list()
            for line in file:
                line = line.strip()
                if "output" in line:
                    split_line = line.split(" output ")
                    points = split_line[0]
                    tour = split_line[1]
                    tour = tour.split(" ")
                    tour = np.array([int(t) for t in tour])
                    tour -= 1
                    ref_tours.append(tour)
                else:
                    points = line
                    load_ref_tours = False
                points = points.split(" ")
                points = np.array(
                    [
                        [float(points[i]), float(points[i + 1])]
                        for i in range(0, len(points), 2)
                    ]
                )
                nodes_coords.append(points)

        if return_list:
            return nodes_coords, ref_tours

        if load_ref_tours:
            try:
                self.ref_tours = np.array(ref_tours)
            except Exception as e:
                message = (
                    "This method does not support instances of different numbers of nodes. "
                    "If you want to read the data, please set ``return_list`` as True. "
                    "Anyway, the data will not be saved in the solver. "
                    "Please convert the data to ``np.ndarray`` externally before calling the solver."
                )
                raise Exception(message) from e

        nodes_coords = np.array(nodes_coords)
        self.ori_points = nodes_coords
        self.points = nodes_coords.astype(np.float32)
        self.check_points_dim()
        self.check_ref_tours_dim()
        if normalize:
            self.normalize_points()

    def from_data(
        self,
        points: Union[list, np.ndarray],
        norm: str = "EUC_2D",
        normalize: bool = False,
    ):
        self.set_norm(norm)
        if points is None:
            return
        self.ori_points = points
        if type(points) == list:
            points = np.array(points)
        self.points = points.astype(np.float32)
        self.check_points_dim()
        if normalize:
            self.normalize_points()

    def read_tours(self, tours: Union[list, np.ndarray]):
        if tours is None:
            return
        if type(tours) == list:
            tours = np.array(tours)
        self.tours = tours
        self.check_tours_dim()

    def read_ref_tours(self, ref_tours: Union[list, np.ndarray]):
        if ref_tours is None:
            return
        if type(ref_tours) == list:
            ref_tours = np.array(ref_tours)
        self.ref_tours = ref_tours
        self.check_ref_tours_dim()

    def read_ref_tours_from_opt_tour(self, file_path: str):
        if not file_path.endswith(".opt.tour"):
            raise ValueError("Invalid file format. Expected a ``.opt.tour`` file.")
        tsp_tour = tsplib95.load(file_path)
        tsp_tour = tsp_tour.tours
        tsp_tour: list
        tsp_tour = tsp_tour[0]
        tsp_tour.append(1)
        self.ref_tours = np.array(tsp_tour) - 1
        self.check_ref_tours_dim()

    def to_tsp(
        self,
        save_dir: str,
        filename: str,
        original: bool = True,
        points: Union[np.ndarray, list] = None,
        norm: str = None,
        normalize: bool = False,
    ):
        # prepare
        self.from_data(points, norm, normalize)
        if filename.endswith(".tsp"):
            filename = os.path.basename(filename)
        self.check_points_not_none()
        self.check_tours_not_none()
        points = self.ori_points if original else self.points

        # makedirs
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        # write
        for idx in range(points.shape[0]):
            save_path = os.path.join(save_dir, filename + f"-{idx}.tsp")
            with open(save_path, "w") as f:
                f.write(f"NAME: {save_path}\n")
                f.write("TYPE: TSP\n")
                f.write(f"DIMENSION: {self.nodes_num}\n")
                f.write(f"EDGE_WEIGHT_TYPE: {self.norm}\n")
                f.write("NODE_COORD_SECTION\n")
                for i in range(self.nodes_num):
                    x, y = points[idx][i]
                    f.write(f"{i+1} {x} {y}\n")
                f.write("EOF\n")

    def to_opt_tour(
        self,
        save_dir: str,
        filename: str,
        tours: Union[np.ndarray, list] = None,
    ):
        # read and check
        self.read_tours(tours)
        if filename.endswith(".opt.tour"):
            filename = filename.replace(".opt.tour", "")
        self.check_tours_not_none()
        tours = self.tours

        # makedirs
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        # write
        for idx in range(tours.shape[0]):
            save_path = os.path.join(save_dir, filename + f"-{idx}.opt.tour")
            with open(save_path, "w") as f:
                f.write(f"NAME: {save_path}\n")
                f.write(f"TYPE: TOUR\n")
                f.write(f"DIMENSION: {self.nodes_num}\n")
                f.write(f"TOUR_SECTION\n")
                for i in range(self.nodes_num):
                    f.write(f"{tours[idx][i]}\n")
                f.write(f"-1\n")
                f.write(f"EOF\n")

    def to_txt(
        self,
        filename: str = "example.txt",
        original: bool = True,
        points: Union[np.ndarray, list] = None,
        tours: Union[np.ndarray, list] = None,
        norm: str = None,
        normalize: bool = False,
    ):
        # read and check
        self.from_data(points, norm, normalize)
        self.read_tours(tours)
        if self.tours is None:
            raise ValueError(
                "``tours`` cannot be None, please use method 'solve' to get solutions."
            )
        self.check_points_not_none()
        self.check_tours_not_none()
        points = self.ori_points if original else self.points
        tours = self.tours

        # write
        with open(filename, "w") as f:
            for node_coordes, tour in zip(points, tours):
                f.write(" ".join(str(x) + str(" ") + str(y) for x, y in node_coordes))
                f.write(str(" ") + str("output") + str(" "))
                f.write(str(" ").join(str(node_idx + 1) for node_idx in tour))
                f.write("\n")
            f.close()

    def evaluate(
        self,
        original: bool = True,
        points: Union[np.ndarray, list] = None,
        tours: Union[np.ndarray, list] = None,
        ref_tours: Union[np.ndarray, list] = None,
        norm: str = None,
        normalize: bool = False,
        calculate_gap: bool = False,
    ):
        # read and check
        self.from_data(points, norm, normalize)
        self.read_tours(tours)
        self.read_ref_tours(ref_tours)
        self.check_points_not_none()
        self.check_tours_not_none()
        self.check_ref_tours_not_none()
        points = self.ori_points if original else self.points
        tours = self.tours
        ref_tours = self.ref_tours

        # prepare for evaluate
        tours_cost_list = list()
        samples = points.shape[0]
        if calculate_gap:
            ref_tours_cost_list = list()
            gap_list = list()

        # deal with different situation
        if tours.shape[0] != samples:
            # a problem has more than one solved tour
            tours = tours.reshape(samples, -1, tours.shape[-1])
            for idx in range(samples):
                evaluator = TSPEvaluator(points[idx], self.norm)
                solved_tours = tours[idx]
                solved_costs = list()
                for tour in solved_tours:
                    solved_costs.append(evaluator.evaluate(tour))
                solved_cost = np.min(solved_costs)
                tours_cost_list.append(solved_cost)
                if calculate_gap:
                    ref_cost = evaluator.evaluate(ref_tours[idx])
                    ref_tours_cost_list.append(ref_cost)
                    gap = (solved_cost - ref_cost) / ref_cost * 100
                    gap_list.append(gap)
        else:
            # a problem only one solved tour
            for idx in range(samples):
                evaluator = TSPEvaluator(points[idx], self.norm)
                solved_tour = tours[idx]
                solved_cost = evaluator.evaluate(solved_tour)
                tours_cost_list.append(solved_cost)
                if calculate_gap:
                    ref_cost = evaluator.evaluate(ref_tours[idx])
                    ref_tours_cost_list.append(ref_cost)
                    gap = (solved_cost - ref_cost) / ref_cost * 100
                    gap_list.append(gap)

        # calculate average cost/gap & std
        tours_costs = np.array(tours_cost_list)
        if calculate_gap:
            ref_costs = np.array(ref_tours_cost_list)
            gaps = np.array(gap_list)
        costs_avg = np.average(tours_costs)
        if calculate_gap:
            ref_costs_avg = np.average(ref_costs)
            gap_avg = np.sum(gaps) / samples
            gap_std = np.std(gaps)
            return costs_avg, ref_costs_avg, gap_avg, gap_std
        else:
            return costs_avg

    def solve(
        self,
        points: Union[np.ndarray, list] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        num_threads: int = 1,
        show_time: bool = False,
        **kwargs,
    ) -> np.ndarray:
        raise NotImplementedError(
            "The ``solve`` function is required to implemented in subclasses."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import re
from setuptools import find_packages, setup
from wheel.bdist_wheel import (
    bdist_wheel as _bdist_wheel,
    get_platform,
    get_abi_tag,
    tags,
)


def get_property(prop, project):
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        open(project + "/__init__.py").read(),
    )
    return result.group(1)


# Package meta-data.
NAME = "ml4co-kit"
PACKAGE_NAME = "ml4co_kit"
DESCRIPTION = "ml4co-kit provides convenient dataset generators for the combinatorial optimization problem"
URL = "https://github.com/Thinklab-SJTU/ML4CO-Kit"
AUTHOR = get_property("__author__", PACKAGE_NAME)
VERSION = get_property("__version__", PACKAGE_NAME)
REQUIRED = [
    "numpy>=1.24.4",
    "networkx>=2.8.8",
    "tsplib95>=0.7.1",
    "tqdm>=4.66.1",
    "requests>=2.31.0",
    "pulp>=2.8.0",
    "pandas>=2.0.0",
    "scipy>=1.10.1",
    "aiohttp>=3.9.3",
]

EXTRAS = {}

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class BdistWheelCommand(_bdist_wheel):
    def run(self):
        os.system("python publish/compile_solvers.py")
        super().run()

    def get_tag(self):
        if self.plat_name and not self.plat_name.startswith("macosx"):
            plat_name = self.plat_name
        else:
            plat_name = get_platform(self.bdist_dir)
        if plat_name in ("linux-x86_64", "linux_x86_64"):
            plat_name = "manylinux2014_x86_64"
        plat_name = (
            plat_name.lower().replace("-", "_").replace(".", "_").replace(" ", "_")
        )
        impl_name = tags.interpreter_name()
        impl_ver = tags.interpreter_version()
        impl = impl_name + impl_ver
        abi_tag = str(get_abi_tag()).lower()
        tag = (impl, abi_tag, plat_name)
        return tag


setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_data={NAME: ["**"], "docs": ["**"]},
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="Mulan PSL v2",
    python_requires=">=3.8",
    classifiers=[
        "License :: OSI Approved :: Mulan Permissive Software License v2 (MulanPSL-2.0)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: Unix",
        "Environment :: Console",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    # $ setup.py publish support.
    cmdclass={
        "bdist_wheel": BdistWheelCommand,
    },
)



import os
from ml4co_kit.solver.mis.base import MISSolver
from ml4co_kit.data.mis.satlib_original import SATLIBOriDataset
from ml4co_kit.utils.mis_utils import cnf_folder_to_gpickle_folder


class SATLIBEvaluator:
    def __init__(self, test_folder: str = "dataset/satlib_original/test_files") -> None:
        self.dataset = SATLIBOriDataset()
        self.test_folder = test_folder
        gpickle_root_foler = test_folder + "_gpickle"
        cnf_folder_to_gpickle_folder(
            cnf_folder=test_folder,
            gpickle_foler=gpickle_root_foler
        )
        self.gpickle_foler = os.path.join(gpickle_root_foler, "mis_graph")
        self.ref_solution_path = os.path.join(gpickle_root_foler, "ref_solution.txt")

    def evaluate(self, solver: MISSolver, **solver_args):
        solver.solve(self.gpickle_foler, **solver_args)
        solver.from_gpickle_folder(self.gpickle_foler, ref=False)
        solver.read_ref_sel_nodes_num_from_txt(self.ref_solution_path)
        return solver.evaluate(calculate_gap=True)



import os
import time
import uuid
import numpy as np
from tqdm import tqdm
from multiprocessing import Pool
from typing import Union
from .pyconcorde import TSPConSolver
from .base import TSPSolver


class TSPConcordeSolver(TSPSolver):
    def __init__(
        self,
        scale: int = 1e6,
    ):
        """
        TSPConcordeSolver
        Args:
            scale (int, optional):
                The scale factor for coordinates in the Concorde solver.
        """
        super(TSPConcordeSolver, self).__init__(solver_type="concorde", scale=scale)

    def _solve(self, nodes_coord: np.ndarray, name: str) -> np.ndarray:
        solver = TSPConSolver.from_data(
            xs=nodes_coord[:, 0] * self.scale,
            ys=nodes_coord[:, 1] * self.scale,
            norm=self.norm,
            name=name,
        )
        solution = solver.solve(verbose=False, name=name)
        tour = solution.tour
        return tour

    def solve(
        self,
        points: Union[np.ndarray, list] = None,
        norm: str = "EUC_2D",
        normalize: bool = False,
        num_threads: int = 1,
        show_time: bool = False,
    ) -> np.ndarray:
        # prepare
        self.from_data(points, norm, normalize)
        start_time = time.time()

        # solve
        tours = list()
        p_shape = self.points.shape
        num_points = p_shape[0]
        if num_threads == 1:
            if show_time:
                for idx in tqdm(range(num_points), desc="Solving TSP Using Concorde"):
                    name = uuid.uuid4().hex
                    tours.append(self._solve(self.points[idx], name))
                    self.clear_tmp_files(name)
            else:
                for idx in range(num_points):
                    name = uuid.uuid4().hex
                    tours.append(self._solve(self.points[idx], name))
                    self.clear_tmp_files(name)
        else:
            batch_points = self.points.reshape(
                -1, num_threads, p_shape[-2], p_shape[-1]
            )
            name_list = list()
            if show_time:
                for idx in tqdm(
                    range(num_points // num_threads), desc="Solving TSP Using Concorde"
                ):
                    for _ in range(num_threads):
                        name_list.append(uuid.uuid4().hex)
                    with Pool(num_threads) as p1:
                        name = uuid.uuid4().hex
                        cur_tours = p1.starmap(
                            self._solve,
                            [
                                (batch_points[idx][inner_idx], name)
                                for inner_idx, name in zip(
                                    range(num_threads), name_list
                                )
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)
                    for name in name_list:
                        self.clear_tmp_files(name)
            else:
                for idx in range(num_points // num_threads):
                    for _ in range(num_threads):
                        name_list.append(uuid.uuid4().hex)
                    with Pool(num_threads) as p1:
                        name = uuid.uuid4().hex
                        cur_tours = p1.starmap(
                            self._solve,
                            [
                                (batch_points[idx][inner_idx], name)
                                for inner_idx, name in zip(
                                    range(num_threads), name_list
                                )
                            ],
                        )
                    for tour in cur_tours:
                        tours.append(tour)
                    for name in name_list:
                        self.clear_tmp_files(name)

        # format
        tours = np.array(tours)
        zeros = np.zeros((tours.shape[0], 1))
        tours = np.append(tours, zeros, axis=1).astype(np.int32)
        if tours.ndim == 2 and tours.shape[0] == 1:
            tours = tours[0]
        self.read_tours(tours)
        end_time = time.time()
        if show_time:
            print(f"Use Time: {end_time - start_time}")
        return tours

    def clear_tmp_files(self, name):
        real_name = name[0:9]
        # tmp file
        sol_filename = f"{real_name}.sol"
        Osol_filename = f"O{real_name}.sol"
        res_filename = f"{real_name}.res"
        Ores_filename = f"O{real_name}.res"
        sav_filename = f"{real_name}.sav"
        Osav_filename = f"O{real_name}.sav"
        pul_filename = f"{real_name}.pul"
        Opul_filename = f"O{real_name}.pul"
        filelist = [
            sol_filename,
            Osol_filename,
            res_filename,
            Ores_filename,
            sav_filename,
            Osav_filename,
            pul_filename,
            Opul_filename,
        ]
        # intermediate file
        for i in range(100):
            filelist.append("{}.{:03d}".format(name[0:8], i + 1))
        # delete
        for file in filelist:
            if os.path.exists(file):
                os.remove(file)



import os
import sys

root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root_folder)
from ml4co_kit.draw.tsp import draw_tsp_solution, draw_tsp_problem
from ml4co_kit.draw.mis import draw_mis_solution, draw_mis_problem

from ml4co_kit.solver import TSPConcordeSolver, KaMISSolver


def test_draw_tsp():
    solver = TSPConcordeSolver(scale=100)
    solver.from_tsp("tests/draw_test/eil101.tsp")
    solver.solve()
    draw_tsp_problem(
        save_path="tests/draw_test/eil101_problem.png",
        points=solver.ori_points,
    )
    draw_tsp_solution(
        save_path="tests/draw_test/eil101_solution.png",
        points=solver.ori_points,
        tours=solver.tours,
    )


def test_mis_tsp():
    # use KaMISSolver to solve the problem
    mis_solver = KaMISSolver()
    mis_solver.solve(src="tests/draw_test/mis_example")

    # draw
    draw_mis_problem(
        save_path="tests/draw_test/mis_problem.png", 
        gpickle_path="tests/draw_test/mis_example/mis_example.gpickle"
    )
    draw_mis_solution(
        save_path="tests/draw_test/mis_solution.png",
        gpickle_path="tests/draw_test/mis_example/mis_example.gpickle",
        result_path="tests/draw_test/mis_example/solve/mis_example_unweighted.result"
    )
    
    
if __name__ == "__main__":
    test_draw_tsp()
    test_mis_tsp()
